# -*- coding: utf-8 -*-
"sets the options of the configuration objects Config object itself"
# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from typing import Union, Set
from itertools import chain
from .error import PropertiesOptionError, ConstError, ConfigError, LeadershipError
from .i18n import _


# If cache and expire is enable, time before cache is expired.
# This delay start first time value/setting is set in cache, even if
# user access several time to value/setting
EXPIRATION_TIME = 5

#List of default properties (you can add new one if needed).
#
#For common properties and personalise properties, if a propery is set for
#an Option and for the Config together, Setting raise a PropertiesOptionError
#
#* Common properties:
#
#hidden
#    option with this property can only get value in read only mode. This
#    option is not available in read write mode.
#
#disabled
#    option with this property cannot be set/get
#
#frozen
#    cannot set value for option with this properties if 'frozen' is set in
#    config
#
#* Special property:
#
#permissive
#    option with 'permissive' cannot raise PropertiesOptionError for properties
#    set in permissive
#    config with 'permissive', whole option in this config cannot raise
#    PropertiesOptionError for properties set in permissive
#
#mandatory
#    should set value for option with this properties if 'mandatory' is set in
#    config
#    example: 'a', ['a'], [None] are valid
#             None, [] are not valid
#
#empty
#    raise mandatory PropertiesOptionError if multi or leader have empty value
#    example: ['a'] is valid
#             [None] is not valid
#
#unique
#    raise ValueError if a value is set twice or more in a multi Option
#
#* Special Config properties:
#
#cache
#    if set, enable cache settings and values
#
#expire
#    if set, settings and values in cache expire after ``expiration_time``
#
#everything_frozen
#    whole option in config are frozen (even if option have not frozen
#    property)
#
#validator
#    launch validator set by user in option (this property has no effect
#    for internal validator)
#
#warnings
#    display warnings during validation
#
#demoting_error_warning
#    all value errors are convert to warning (ValueErrorWarning)
DEFAULT_PROPERTIES = frozenset(['cache', 'validator', 'warnings'])
SPECIAL_PROPERTIES = {'frozen', 'mandatory', 'empty', 'force_store_value'}

#Config can be in two defaut mode:
#
#read_only
#    you can get all variables not disabled but you cannot set any variables
#    if a value has a callback without any value, callback is launch and value
#    of this variable can change
#    you cannot access to mandatory variable without values
#
#read_write
#    you can get all variables not disabled and not hidden
#    you can set all variables not frozen
RO_APPEND = frozenset(['frozen',
                       'disabled',
                       'validator',
                       'everything_frozen',
                       'mandatory',
                       'empty',
                       'force_store_value',
                       ])
RO_REMOVE = frozenset(['permissive',
                       'hidden',
                       ])
RW_APPEND = frozenset(['frozen',
                       'disabled',
                       'validator',
                       'hidden',
                       'force_store_value',
                       ])
RW_REMOVE = frozenset(['permissive',
                       'everything_frozen',
                       'mandatory',
                       'empty',
                       ])


FORBIDDEN_SET_PROPERTIES = frozenset(['force_store_value'])
FORBIDDEN_SET_PERMISSIVES = frozenset(['force_default_on_freeze',
                                       'force_metaconfig_on_freeze',
                                       'force_store_value',
                                       ])
ALLOWED_LEADER_PROPERTIES = {'empty',
                             'notempty',
                             'notunique',
                             'unique',
                             'force_store_value',
                             'mandatory',
                             'force_default_on_freeze',
                             'force_metaconfig_on_freeze',
                             'frozen',
                             }

static_set = frozenset()


# ____________________________________________________________
class Undefined:
    """Object undefined, means that there is not value
    """
    # pylint: disable=too-few-public-methods
    def __str__(self):  # pragma: no cover
        return 'Undefined'

    __repr__ = __str__


undefined = Undefined()


class OptionBag:
    """Object to store information for an option
    """
    __slots__ = ('option',  # current option
                 'path',
                 'index',
                 'config_bag',
                 'ori_option',  # original option (for example useful for symlinkoption)
                 'properties',  # properties of current option
                 'apply_requires',  # apply requires or not for this option
                 )

    # pylint: disable=too-many-arguments
    def __init__(self,
                 option,
                 index,
                 config_bag,
                 path=None,
                 properties=undefined,
                 ori_option=None,
                 apply_requires=True,
                 ):
        self.index = index
        self.config_bag = config_bag
        self.option = option
        if ori_option is not None:
            self.ori_option = ori_option
        if config_bag is undefined:
            self.path = path
        elif path:
            self.path = path
        elif option:
            self.path = option.impl_getpath()
            context = config_bag.context
            if '.' not in self.path and option == context.get_description():
                self.properties = None
            elif properties is undefined:
                settings = context.get_settings()
                try:
                    self.properties = settings.getproperties(self,
                                                             apply_requires=apply_requires,
                                                             )
                except ConfigError:
                    self.properties = None
        if properties is not undefined:
            self.properties = properties

    def __getattr__(self, key):
        if key == 'ori_option':
            return self.option
        if key == 'apply_requires':
            return True
        return None

    def copy(self):
        """copy OptionBag
        """
        option_bag = OptionBag(None,
                               None,
                               None,
                               )
        for key in self.__slots__:
            setattr(option_bag, key, getattr(self, key))
        return option_bag


class ConfigBag:
    """Object to store information for context
    """
    __slots__ = ('context',  # link to the current context
                 'properties',  # properties for current context
                 'true_properties',  # properties for current context
                 'is_unrestraint',
                 'permissives',  # permissives for current context
                 'expiration_time',  # EXPIRATION_TIME
                 )

    def __init__(self,
                 context,
                 properties: set,
                 permissives: frozenset,
                 **kwargs):
        self.context = context
        self.properties = properties
        self.permissives = permissives
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, key):
        if key == 'true_properties':
            return self.properties
        if key == 'expiration_time':
            self.expiration_time = EXPIRATION_TIME  # pylint: disable=attribute-defined-outside-init
            return self.expiration_time
        if key == 'is_unrestraint':
            return False
        raise KeyError(f'unknown key "{key}" for ConfigBag')  # pragma: no cover

    def nowarnings(self):
        """do not warnings
        """
        self.properties = frozenset(self.properties - {'warnings'})

    def remove_validation(self):
        """do not validate option
        """
        self.properties = frozenset(self.properties - {'validator'})

    def unrestraint(self):
        """do not restraint access to option
        """
        self.is_unrestraint = True  # pylint: disable=attribute-defined-outside-init
        self.true_properties = self.properties  # pylint: disable=attribute-defined-outside-init
        self.properties = frozenset(['cache'])

    def set_permissive(self):
        """set permissive
        """
        self.properties = frozenset(self.properties | {'permissive'})

    def copy(self):
        """copy the config
        """
        kwargs = {}
        for key in self.__slots__:
            kwargs[key] = getattr(self, key)
        return ConfigBag(**kwargs)


# ____________________________________________________________
class _NameSpace:
    """convenient class that emulates a module
    and builds constants (that is, unique names)
    when attribute is added, we cannot delete it
    """

    def __setattr__(self,
                    name,
                    value,
                    ):
        if name in self.__dict__:
            raise ConstError(_("can't rebind {0}").format(name))
        self.__dict__[name] = value

    def __delattr__(self,
                    name,
                    ):
        raise ConstError(_("can't unbind {0}").format(name))


class GroupModule(_NameSpace):
    "emulates a module to manage unique group (OptionDescription) names"
    # pylint: disable=too-few-public-methods
    class GroupType(str):
        """allowed normal group (OptionDescription) names
        *normal* means : groups that are not leader
        """

    class DefaultGroupType(GroupType):
        """groups that are default (typically 'default')"""

    class LeadershipGroupType(GroupType):
        """allowed normal group (OptionDescription) names
        *leadership* means : groups that have the 'leadership' attribute set
        """

    class RootGroupType(GroupType):
        """root means this is the root optiondescription of whole config
        """

    def addgroup(self, name):
        """add a new group type
        """
        setattr(groups, name, groups.GroupType(name))


class OwnerModule(_NameSpace):
    """emulates a module to manage unique owner names.

    owners are living in `Config._value_owners`
    """
    # pylint: disable=too-few-public-methods
    class Owner(str):
        """allowed owner names
        """

    class DefaultOwner(Owner):
        """groups that are default (typically 'default')"""

    def addowner(self, name):
        """
        :param name: the name of the new owner
        """
        setattr(owners, name, owners.Owner(name))


# ____________________________________________________________
# populate groups
groups = GroupModule()

# groups.default: default group set when creating a new optiondescription
groups.default = groups.DefaultGroupType('default')  # pylint: disable=attribute-defined-outside-init

# groups.leadership: leadership group is a special optiondescription, all suboptions should
#                    be multi option and all values should have same length, to find
#                    leader's option, the optiondescription's name should be same than de
#                    leader's option"""
groups.leadership = groups.LeadershipGroupType('leadership')  # pylint: disable=attribute-defined-outside-init

# groups.root: this group is the root optiondescription of whole config
groups.root = groups.RootGroupType('root')  # pylint: disable=attribute-defined-outside-init


# ____________________________________________________________
# populate owners with default attributes
owners = OwnerModule()

# default: is the config owner after init time
owners.default = owners.DefaultOwner('default')  # pylint: disable=attribute-defined-outside-init

# user: is the generic is the generic owner
owners.addowner('user')

#forced: special owner when value is forced
owners.addowner('forced')


forbidden_owners = (owners.default, owners.forced)  # pylint: disable=no-member


# ____________________________________________________________
class Settings:
    "``config.Config()``'s configuration options settings"
    __slots__ = ('_properties',
                 '_permissives',
                 '_permissives',
                 '__weakref__',
                 'ro_append',
                 'ro_remove',
                 'rw_append',
                 'rw_remove',
                 'default_properties',
                 )

    def __init__(self):
        """
        initializer

        :param context: the root config
        :param storage: the storage type
        """
        # generic owner
        self._properties = {}
        self._permissives = {}
        self.default_properties = DEFAULT_PROPERTIES
        self.ro_append = RO_APPEND
        self.ro_remove = RO_REMOVE
        self.rw_append = RW_APPEND
        self.rw_remove = RW_REMOVE

    # ____________________________________________________________
    # get properties and permissive methods

    def get_context_properties(self,
                               cache,
                               ):
        """get context properties
        """
        is_cached, props, _ = cache.getcache(None,
                                             'context_props',
                                             expiration=False,
                                             )
        if not is_cached:
            props = self.get_stored_properties(None,
                                               None,
                                               self.default_properties,
                                               )
            cache.setcache(None,
                           props,
                           type_='properties',
                           )
        return props

    def get_stored_properties(self,
                              path: Union[None, str],
                              index: Union[None, int],
                              default_properties: Set[str],
                              ) -> Set[str]:
        """Get the properties modified by user for a path or index
        """
        if path not in self._properties or index not in self._properties[path]:
            return frozenset(default_properties)
        return self._properties[path][index]

    def getproperties(self,
                      option_bag,
                      apply_requires=True,
                      uncalculated=False,
                      help_property=False,
                      transitive_raise=True,
                      ):
        """get properties
        """
        # pylint: disable=too-many-branches
        option = option_bag.option
        if option.impl_is_symlinkoption():
            option = option.impl_getopt()
        if apply_requires and not uncalculated and not help_property:
            cache = option_bag.config_bag.context.properties_cache
            is_cached, props, validated = cache.getcache(option_bag,  # pylint: disable=unused-variable
                                                         'self_props',
                                                         )
        else:
            is_cached = False
        if not is_cached:
            props = set()
            # if index, get option's properties (without index) too
            p_props = self.get_stored_properties(option_bag.path,
                                                 None,
                                                 option.impl_getproperties(),
                                                 )
            if option_bag.index is not None:
                p_props = chain(p_props,
                                self.get_stored_properties(option_bag.path,
                                                           option_bag.index,
                                                           option.impl_getproperties(),
                                                           )
                                )
            for prop in p_props:
                if uncalculated or isinstance(prop, str):
                    if not help_property:
                        props.add(prop)
                    else:
                        props.add((prop, prop))
                elif apply_requires:
                    try:
                        if not help_property:
                            new_prop = prop.execute(option_bag,
                                                    for_settings=True,
                                                    )
                        else:
                            new_prop = prop.help(option_bag,
                                                 for_settings=True,
                                                 )
                            if isinstance(new_prop, str):
                                new_prop = (new_prop, new_prop)
                        if new_prop is None:
                            continue
                    except ConfigError as err:
                        if transitive_raise:
                            raise err from err
                        continue
                    if (not help_property and not isinstance(new_prop, str)) or \
                            (help_property and not isinstance(new_prop, tuple)):
                        raise ValueError(_('invalid property type {type(new_prop)} for '
                                           '{option_bag.option.impl_getname()} with '
                                           '{prop.function.__name__} function'))
                    if not option.impl_is_optiondescription() and \
                            option.impl_is_leader() and \
                            new_prop not in ALLOWED_LEADER_PROPERTIES:
                        raise LeadershipError(_('leader cannot have "{new_prop}" property'))
                    props.add(new_prop)
            props -= self.getpermissives(option_bag)
            if not uncalculated and apply_requires and \
                    not option_bag.config_bag.is_unrestraint and \
                    not help_property and transitive_raise:
                cache.setcache(option_bag,
                               props,
                               type_='properties',
                               )
        return props

    def get_context_permissives(self):
        """get context permissives
        """
        return self.getpermissives(None)

    def _getpermissives(self,
                        path,
                        index,
                        ):
        if not path in self._permissives:
            ret = frozenset()
        else:
            ret = self._permissives[path].get(index, frozenset())
        return ret

    def getpermissives(self,
                       option_bag,
                       ):
        """get permissive
        """
        if option_bag is None:
            path = None
            index = None
        else:
            opt = option_bag.option
            if opt.impl_is_symlinkoption():
                opt = opt.impl_getopt()
                path = opt.impl_getpath()
            else:
                path = option_bag.path
            index = option_bag.index
        permissives = self._getpermissives(path,
                                           None,
                                           )
        if index is not None:
            option_permissives = self._permissives.get(path, {}).get(index, set())
            permissives = frozenset(option_permissives | permissives)
        return permissives

    #____________________________________________________________
    # set methods
    def set_context_properties(self, properties, context):
        """set context properties
        """
        self._properties.setdefault(None, {})[None] = properties
        context.reset_cache(None)

    def setproperties(self,
                      option_bag,
                      properties,
                      ):
        """save properties for specified path
        (never save properties if same has option properties)
        """
        opt = option_bag.option
        if not opt.impl_is_optiondescription() and opt.impl_is_leader():
            not_allowed_properties = properties - ALLOWED_LEADER_PROPERTIES
            if not_allowed_properties:
                raise LeadershipError(_('leader cannot have "{list(not_allowed_properties)}" '
                                        'property'))
            if ('force_default_on_freeze' in properties or \
                    'force_metaconfig_on_freeze' in properties) and 'frozen' not in properties:
                raise LeadershipError(_('a leader ({opt.impl_get_display_name()}) cannot have '
                                        '"force_default_on_freeze" or '
                                        '"force_metaconfig_on_freeze" property without "frozen"'))
        self._properties.setdefault(option_bag.path, {})[option_bag.index] = properties
        # values too because of follower values could have a PropertiesOptionError has value
        option_bag.config_bag.context.reset_cache(option_bag)
        option_bag.properties = properties

    def set_context_permissives(self,
                                permissives,
                                ):
        """set context permissive
        """
        self.setpermissives(None,
                            permissives,
                            )

    def setpermissives(self,
                       option_bag,
                       permissives,
                       ):
        """
        enables us to put the permissives in the storage

        :param path: the option's path
        :param type: str
        :param opt: if an option object is set, the path is extracted.
                    it is better (faster) to set the path parameter
                    instead of passing a :class:`tiramisu.option.Option()` object.
        """
        if not isinstance(permissives, frozenset):
            raise TypeError(_('permissive must be a frozenset'))
        if option_bag is not None:
            path = option_bag.path
            index = option_bag.index
        else:
            path = None
            index = None
        forbidden_permissives = FORBIDDEN_SET_PERMISSIVES & permissives
        if forbidden_permissives:
            raise ConfigError(_('cannot add those permissives: {0}').format(
                ' '.join(forbidden_permissives)))
        self._permissives.setdefault(path, {})[index] = permissives
        if option_bag is not None:
            option_bag.config_bag.context.reset_cache(option_bag)

    #____________________________________________________________
    # reset methods
    def _get_path_index_config_option(self,
                                      bag: Union[OptionBag, ConfigBag],
                                      msg: str,
                                      ):
        if isinstance(bag, ConfigBag):
            path = None
            index = None
            config_bag = bag
            option_bag = None
        else:
            assert not bag.option.impl_is_symlinkoption(), \
                    _(msg).format(bag.option.impl_get_display_name())
            path = bag.path
            index = bag.index
            config_bag = bag.config_bag
            option_bag = bag
        return path, index, config_bag, option_bag

    def reset(self,
              bag: Union[OptionBag, ConfigBag],
              ):
        """reset property
        """
        path, index, config_bag, option_bag = \
                self._get_path_index_config_option(bag,
                                                   "can't reset properties to "
                                                   "the symlinkoption \"{}\"",
                                                   )
        if path in self._properties and index in self._properties[path]:
            del self._properties[path][index]
        config_bag.context.reset_cache(option_bag)

    def reset_permissives(self,
                          bag: Union[OptionBag, ConfigBag],
                          ):
        """reset permission
        """
        path, index, config_bag, option_bag = \
                self._get_path_index_config_option(bag,
                                                   "can't reset permissives to "
                                                   "the symlinkoption \"{}\"",
                                                   )
        if path in self._permissives and index in self._permissives[path]:
            del self._permissives[path][index]
        config_bag.context.reset_cache(option_bag)

    #____________________________________________________________
    # validate properties
    def calc_raises_properties(self,
                               option_bag,
                               apply_requires=True,
                               uncalculated=False,
                               transitive_raise=True,
                               ):
        """raise if needed
        """
        if not uncalculated and apply_requires and option_bag.properties is not None:
            option_properties = option_bag.properties
        else:
            option_properties = self.getproperties(option_bag,
                                                   apply_requires=apply_requires,
                                                   uncalculated=uncalculated,
                                                   transitive_raise=transitive_raise,
                                                   )
        return self._calc_raises_properties(option_bag.config_bag.properties,
                                            option_bag.config_bag.permissives,
                                            option_properties,
                                            )

    def _calc_raises_properties(self,
                                context_properties,
                                context_permissives,
                                option_properties,
                                ):
        raises_properties = context_properties - SPECIAL_PROPERTIES
        # remove global permissive properties
        if raises_properties and 'permissive' in raises_properties:
            raises_properties -= context_permissives
        properties = option_properties & raises_properties
        # at this point it should not remain any property for the option
        return properties

    def validate_properties(self,
                            option_bag,
                            need_help=True,
                            ):
        """check properties
        """
        config_properties = option_bag.config_bag.properties
        if not config_properties or config_properties == frozenset(['cache']):
            # if no global property
            return
        for transitive_raise in [False, True]:
            properties = self.calc_raises_properties(option_bag,
                                                     transitive_raise=transitive_raise,
                                                     )
            if properties != frozenset():
                if need_help:
                    help_properties = dict(self.getproperties(option_bag,
                                                              help_property=True,
                                                              transitive_raise=transitive_raise,
                                                              ))
                    calc_properties = []
                    for property_ in self._calc_raises_properties(option_bag.config_bag.properties,
                                                                  option_bag.config_bag.permissives,
                                                                  set(help_properties.keys()),
                                                                  ):
                        calc_properties.append(help_properties[property_])
                    calc_properties = frozenset(calc_properties)
                else:
                    calc_properties = properties
                raise PropertiesOptionError(option_bag,
                                            properties,
                                            self,
                                            help_properties=calc_properties,
                                            )

    def validate_mandatory(self,
                           value,
                           option_bag,
                           ):
        """verify if option is mandatory without value
        """
        if 'mandatory' not in option_bag.config_bag.properties:
            return
        values = option_bag.config_bag.context.get_values()
        if not ('permissive' in option_bag.config_bag.properties and
                'mandatory' in option_bag.config_bag.permissives) and \
                'mandatory' in option_bag.properties and values.isempty(option_bag,
                                                                        value,
                                                                        False,
                                                                        ):
            raise PropertiesOptionError(option_bag,
                                        ['mandatory'],
                                        self,
                                        )
        if 'empty' in option_bag.properties and values.isempty(option_bag,
                                                               value,
                                                               True,
                                                               ):
            raise PropertiesOptionError(option_bag,
                                        ['empty'],
                                        self,
                                        )

    def validate_frozen(self,
                        option_bag,
                        ):
        """verify if option is frozen
        """
        if option_bag.config_bag.properties and \
                ('everything_frozen' in option_bag.config_bag.properties or
                 ('frozen' in option_bag.config_bag.properties and \
                         'frozen' in option_bag.properties)) and \
                not (('permissive' in option_bag.config_bag.properties) and
                     'frozen' in option_bag.config_bag.permissives):
            raise PropertiesOptionError(option_bag,
                                        ['frozen'],
                                        self,
                                        )
        return False
    #____________________________________________________________
    # read only/read write

    def _read(self,
              remove,
              append,
              config_bag,
              ):
        props = self.get_stored_properties(None,
                                           None,
                                           self.default_properties,
                                           )
        modified = False
        if remove & props:
            props = props - remove
            modified = True
        if append & props != append:
            props = props | append
            modified = True
        if modified:
            self.set_context_properties(frozenset(props),
                                        config_bag.context,
                                        )

    def read_only(self,
                  config_bag,
                  ):
        "convenience method to freeze, hide and disable"
        self._read(self.ro_remove,
                   self.ro_append,
                   config_bag,
                   )

    def read_write(self,
                   config_bag,
                   ):
        "convenience method to freeze, hide and disable"
        self._read(self.rw_remove,
                   self.rw_append,
                   config_bag,
                   )



# Copyright (C) 2018-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""some functions to validates or calculates value
"""
from typing import Any, List, Optional
from operator import add, mul, sub, truediv
from ipaddress import ip_address, ip_interface, ip_network
from .i18n import _
from .setting import undefined
from .error import display_list


FUNCTION_WAITING_FOR_DICT = []


def function_waiting_for_dict(function):
    """functions (calculation or validation) receive by default only the value of other options
    all functions declared with this function recieve a dict with option informations
    (value, name, ...)
    """
    name = function.__name__
    if name not in FUNCTION_WAITING_FOR_DICT:
        FUNCTION_WAITING_FOR_DICT.append(name)
    return function


@function_waiting_for_dict
def valid_network_netmask(network: dict,
                          netmask: dict,
                          ):
    """
    validates if network and netmask are coherent
    this validator must be set to netmask option
    """
    if None in [network['value'], netmask['value']]:
        return
    try:
        ip_network(f'{network["value"]}/{netmask["value"]}')
    except ValueError as err:
        raise ValueError(_(f'network "{network["value"]}" ({network["name"]}) does not match '
                           'with this netmask')) from err


@function_waiting_for_dict
def valid_ip_netmask(ip: dict,  # pylint: disable=invalid-name
                     netmask: dict,
                     ):
    """validates if ip and netmask are coherent
    this validator must be set to netmask option
    """
    if None in [ip['value'], netmask['value']]:
        return
    ip_netmask = ip_interface(f'{ip["value"]}/{netmask["value"]}')
    if ip_netmask.ip == ip_netmask.network.network_address:
        msg = _(f'IP "{ip["value"]}" ({ip["name"]}) with this netmask is '
                'in fact a network address')
        raise ValueError(msg)
    if ip_netmask.ip == ip_netmask.network.broadcast_address:
        msg = _(f'IP "{ip["value"]}" ({ip["name"]}) with this netmask is '
                'in fact a broacast address')
        raise ValueError(msg)


@function_waiting_for_dict
def valid_broadcast(network: dict,
                    netmask: dict,
                    broadcast: dict,
                    ):
    """validates if the broadcast is coherent with network and netmask
    """
    if None in [network['value'], netmask['value'], broadcast['value']]:
        return
    if ip_network(f'{network["value"]}/{netmask["value"]}').broadcast_address != \
            ip_address(broadcast['value']):
        msg = _(f'broadcast invalid with network {network["value"]} ({network["name"]}) '
                f'and netmask {netmask["value"]} ({netmask["name"]})')
        raise ValueError(msg)


@function_waiting_for_dict
def valid_in_network(ip: dict,  # pylint: disable=invalid-name
                     network: dict,
                     netmask=Optional[dict],
                     ):
    """validates if an IP is in a network
    this validator must be set to ip option
    """
    if None in [ip['value'], network['value']]:
        return
    if '/' in network['value']:
        # it's a CIDR network
        network_value = network['value']
    else:
        if netmask is None or netmask['value'] is None:
            return
        network_value = f'{network["value"]}/{netmask["value"]}'
    network_obj = ip_network(network_value)
    ip_netmask = ip_interface(f'{ip["value"]}/{network_obj.netmask}')
    if ip_netmask not in network_obj:
        if netmask is None:
            msg = _('this IP is not in network {network["value"]} ({network["name"]})')
        else:
            msg = _('this IP is not in network {network["value"]} ({network["name"]}) '
                    'with netmask {netmask["value"]} ({netmask["name"]})')
        raise ValueError(msg)
    # test if ip is not network/broadcast IP
    if ip_netmask.ip == ip_netmask.network.network_address:
        msg = _(f'this IP with the network {network["value"]} ({network["value"]} '
                'is in fact a network address')
        raise ValueError(msg)
    if ip_netmask.ip == ip_netmask.network.broadcast_address:
        msg = _(f'this IP with the network {network["value"]} ({network["value"]} '
                'is in fact a broadcast address')
        raise ValueError(msg)


@function_waiting_for_dict
def valid_not_equal(*values):
    """valid that two options have not same value
    """
    equal = set()
    for val in values[1:]:
        if 'propertyerror' in val:
            continue
        if values[0]['value'] == val['value'] is not None:
            equal.add(val['name'])
    if not equal:
        return
    msg = _(f'value is identical to {display_list(list(equal), add_quote=True)}')
    raise ValueError(msg)


class CalcValue:
    """class to calc_value with different functions
    """
    # pylint: disable=too-many-instance-attributes
    def __call__(self,
                 *args: List[Any],
                 multi: bool=False,
                 default: Any=undefined,
                 condition: Any=undefined,
                 no_condition_is_invalid: bool=False,
                 expected: Any=undefined,
                 condition_operator: str='AND',
                 reverse_condition: bool=False,
                 allow_none: bool=False,
                 remove_duplicate_value: bool=False,
                 join: Optional[str]=None,
                 min_args_len: Optional[int]=None,
                 operator: Optional[str]=None,
                 index: Optional[int]=None,
                 **kwargs) -> Any:
        # pylint: disable=too-many-statements,too-many-branches,too-many-nested-blocks,too-many-locals
        """calculate value
        :param args: list of value
        :param multi: value returns must be a list of value
        :param default: default value if condition is not matched or if args is empty
                        if there is more than one default value, set default_0, default_1, ...
        :param condition: test if condition is equal to expected value
                          if there is more than one condition, set condition_0, condition_1, ...
        :param expected: value expected for all conditions
                         if expected value is different between condition, set expected_0,
                         expected_1, ...
        :param no_condition_is_invalid: if no condition and not condition_0, condition_1, ... (for
                                        example if option is disabled) consider that condition not
                                        matching
        :param condition_operator: OR or AND operator for condition
        :param allow_none: if False, do not return list in None is present in list
        :param remove_duplicate_value: if True, remote duplicated value
        :param join: join all args with specified characters
        :param min_args_len: if number of arguments is smaller than this value, return default value
        :param operator: 'add', 'mul', 'div' or 'sub' all args (args must be integer value)
        :param index: index for follower

        examples:
        * you want to copy value from an option to an other option:
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, \
        ...                      Params, ParamOption
        >>> val1 = StrOption('val1', '', 'val1')
        >>> val2 = StrOption('val2', '', callback=calc_value,
        ...                  callback_params=Params(ParamOption(val1)))
        >>> od = OptionDescription('root', '', [val1, val2])
        >>> cfg = Config(od)
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val1'}

        * you want to copy values from two options in one multi option
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = StrOption('val1', "", 'val1')
        >>> val2 = StrOption('val2', "", 'val2')
        >>> val3 = StrOption('val3', "", multi=True, callback=calc_value,
        ...                  callback_params=Params((ParamOption(val1), ParamOption(val2)),
        ...                  multi=ParamValue(True)))
        >>> od = OptionDescription('root', '', [val1, val2, val3])
        >>> cfg = Config(od)
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val2', 'val3': ['val1', 'val2']}

        * you want to copy a value from an option if it not disabled, otherwise set 'default_value'
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = StrOption('val1', '', 'val1')
        >>> val2 = StrOption('val2', '', callback=calc_value,
        ...                  callback_params=Params(ParamOption(val1, True),
        ...                  default=ParamValue('default_value')))
        >>> od = OptionDescription('root', '', [val1, val2])
        >>> cfg = Config(od)
        >>> cfg.property.read_write()
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val1'}
        >>> cfg.option('val1').property.add('disabled')
        >>> cfg.value.dict()
        {'val2': 'default_value'}

        * you want to copy value from an option if an other is True, otherwise set 'default_value'
        >>> from tiramisu import calc_value, BoolOption, StrOption, OptionDescription, Config, \
        ...                      Params, ParamOption, ParamValue
        >>> boolean = BoolOption('boolean', '', True)
        >>> val1 = StrOption('val1', '', 'val1')
        >>> val2 = StrOption('val2', '', callback=calc_value,
        ...                  callback_params=Params(ParamOption(val1, True),
        ...                                         default=ParamValue('default_value'),
        ...                                         condition=ParamOption(boolean),
        ...                                         expected=ParamValue(True)))
        >>> od = OptionDescription('root', '', [boolean, val1, val2])
        >>> cfg = Config(od)
        >>> cfg.property.read_write()
        >>> cfg.value.dict()
        {'boolean': True, 'val1': 'val1', 'val2': 'val1'}
        >>> cfg.option('boolean').value.set(False)
        >>> cfg.value.dict()
        {'boolean': False, 'val1': 'val1', 'val2': 'default_value'}

        * you want to copy option even if None is present
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = StrOption('val1', "", 'val1')
        >>> val2 = StrOption('val2', "")
        >>> val3 = StrOption('val3', "", multi=True, callback=calc_value,
        ...                  callback_params=Params((ParamOption(val1), ParamOption(val2)),
        ...                  multi=ParamValue(True), allow_none=ParamValue(True)))
        >>> od = OptionDescription('root', '', [val1, val2, val3])
        >>> cfg = Config(od)
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': None, 'val3': ['val1', None]}

        * you want uniq value
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = StrOption('val1', "", 'val1')
        >>> val2 = StrOption('val2', "", 'val1')
        >>> val3 = StrOption('val3', "", multi=True, callback=calc_value,
        ...                  callback_params=Params((ParamOption(val1), ParamOption(val2)),
        ...                  multi=ParamValue(True), remove_duplicate_value=ParamValue(True)))
        >>> od = OptionDescription('root', '', [val1, val2, val3])
        >>> cfg = Config(od)
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val1', 'val3': ['val1']}

        * you want to join two values with '.'
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = StrOption('val1', "", 'val1')
        >>> val2 = StrOption('val2', "", 'val2')
        >>> val3 = StrOption('val3', "", callback=calc_value,
        ...                  callback_params=Params((ParamOption(val1),
        ...                  ParamOption(val2)), join=ParamValue('.')))
        >>> od = OptionDescription('root', '', [val1, val2, val3])
        >>> cfg = Config(od)
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val2', 'val3': 'val1.val2'}

        * you want join three values, only if almost three values are set
        >>> from tiramisu import calc_value, StrOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = StrOption('val1', "", 'val1')
        >>> val2 = StrOption('val2', "", 'val2')
        >>> val3 = StrOption('val3', "", 'val3')
        >>> val4 = StrOption('val4', "", callback=calc_value,
        ...                              callback_params=Params((ParamOption(val1),
        ...                                                      ParamOption(val2),
        ...                                                      ParamOption(val3, True)),
        ...                              join=ParamValue('.'), min_args_len=ParamValue(3)))
        >>> od = OptionDescription('root', '', [val1, val2, val3, val4])
        >>> cfg = Config(od)
        >>> cfg.property.read_write()
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val2', 'val3': 'val3', 'val4': 'val1.val2.val3'}
        >>> cfg.option('val3').property.add('disabled')
        >>> cfg.value.dict()
        {'val1': 'val1', 'val2': 'val2', 'val4': ''}

        * you want to add all values
        >>> from tiramisu import calc_value, IntOption, OptionDescription, Config, Params, \
        ...                      ParamOption, ParamValue
        >>> val1 = IntOption('val1', "", 1)
        >>> val2 = IntOption('val2', "", 2)
        >>> val3 = IntOption('val3', "", callback=calc_value,
        ...                              callback_params=Params((ParamOption(val1),
                                                                 ParamOption(val2)),
        ...                              operator=ParamValue('add')))
        >>> od = OptionDescription('root', '', [val1, val2, val3])
        >>> cfg = Config(od)
        >>> cfg.value.dict()
        {'val1': 1, 'val2': 2, 'val3': 3}

        """
        # pylint: disable=attribute-defined-outside-init
        self.args = args
        self.condition = condition
        self.expected = expected
        self.condition_operator = condition_operator
        self.reverse_condition = reverse_condition
        self.kwargs = kwargs
        self.no_condition_is_invalid = no_condition_is_invalid  # pylint: disable=attribute-defined-outside-init
        value = self.get_value(default,
                               min_args_len,
                               )
        if not multi:
            if join is not None:
                if None not in value:
                    value = join.join(value)
                else:
                    value = None
            elif value and operator:
                new_value = value[0]
                oper = {'mul': mul,
                        'add': add,
                        'div': truediv,
                        'sub': sub,
                        }[operator]
                for val in value[1:]:
                    new_value = oper(new_value, val)
                value = new_value
            elif value == []:
                value = None
            else:
                value = value[0]
                if isinstance(value, list) and index is not None:
                    if len(value) > index:
                        value = value[index]
                    else:
                        value = None
        else:
            if join is not None:
                if None not in value:
                    length_val = None
                    for val in value:
                        if isinstance(val, list):
                            if None in val:
                                length_val = None
                                break
                            lval = len(val)
                            if length_val is not None and length_val != lval:
                                msg = _('unexpected value in calc_value with join attribute '
                                      f'"{val}" with invalid length "{length_val}"')
                                raise ValueError(msg)
                            length_val = lval
                    new_value = []
                    if length_val is not None:
                        for idx in range(length_val):
                            idx_val = []
                            for val in value:
                                if isinstance(val, list):
                                    idx_val.append(val[idx])
                                else:
                                    idx_val.append(val)
                            new_value.append(join.join(idx_val))
                    value = new_value
                else:
                    value = []
            elif None in value and not allow_none:
                value = []
            if remove_duplicate_value:
                new_value = []
                for val in value:
                    if val not in new_value:
                        new_value.append(val)
                value = new_value
        return value

    def value_from_kwargs(self,
                          value: Any,
                          pattern: str,
                          to_dict: bool=False,
                          empty_test=undefined) -> Any:
        """get value from kwargs
        """
        # pylint: disable=too-many-branches
        # if value attribute exist return it's value
        # otherwise pattern_0, pattern_1, ...
        # otherwise undefined
        if value is not empty_test:
            if to_dict == 'all':
                returns = {None: value}
            else:
                returns = value
        else:
            kwargs_matches = {}
            len_pattern = len(pattern)
            for key, pattern_value in self.kwargs.items():
                if key.startswith(pattern):
                    index = int(key[len_pattern:])
                    if isinstance(pattern_value, dict):
                        pattern_value = pattern_value['value']
                    kwargs_matches[index] = pattern_value
            if not kwargs_matches:
                returns = undefined
            else:
                keys = sorted(kwargs_matches)
                if to_dict:
                    returns = {}
                else:
                    returns = []
                for key in keys:
                    if to_dict:
                        returns[key] = kwargs_matches[key]
                    else:
                        returns.append(kwargs_matches[key])
        return returns

    def is_condition_matches(self,
                             condition_value,
                             ):
        """verify the condition
        """
        # pylint: disable=too-many-branches
        calculated_conditions = self.value_from_kwargs(condition_value,
                                                       'condition_',
                                                       to_dict='all',
                                                       )
        if calculated_conditions is undefined:
            is_matches = not self.no_condition_is_invalid
        else:
            is_matches = None
            calculated_expected = self.value_from_kwargs(self.expected,
                                                         'expected_',
                                                         to_dict=True,
                                                         )
            calculated_reverse = self.value_from_kwargs(self.reverse_condition,
                                                        'reverse_condition_',
                                                        to_dict=True,
                                                        empty_test=False,
                                                        )
            for idx, calculated_condition in calculated_conditions.items():
                if isinstance(calculated_expected, dict):
                    if idx is not None:
                        if isinstance(calculated_expected[idx], list):
                            current_matches = calculated_condition in calculated_expected[idx]
                        else:
                            current_matches = calculated_condition == calculated_expected[idx]
                    else:
                        current_matches = calculated_condition in calculated_expected.values()
                else:
                    current_matches = calculated_condition == calculated_expected
                if isinstance(calculated_reverse, dict) and idx in calculated_reverse:
                    reverse_condition = calculated_reverse[idx]
                else:
                    reverse_condition = False
                if is_matches is None:
                    is_matches = current_matches
                if self.condition_operator == 'AND':
                    is_matches = is_matches and current_matches
                    if reverse_condition:
                        is_matches = not is_matches
                    if not is_matches:
                        break
                elif self.condition_operator == 'OR':
                    is_matches = is_matches or current_matches
                    if reverse_condition:
                        is_matches = not is_matches
                    if is_matches:
                        break
                else:
                    msg = _(f'unexpected {self.condition_operator} condition_operator '
                            'in calc_value')
                    raise ValueError(msg)
            is_matches = is_matches and not self.reverse_condition \
                         or not is_matches and self.reverse_condition
        return is_matches

    def get_value(self,
                  default,
                  min_args_len,
                  ):
        """get the value from arguments
        """
        # retrieve the condition
        if isinstance(self.condition, dict):
            if 'value' in self.condition:
                condition_value = self.condition['value']
            else:
                condition_value = undefined
        else:
            condition_value = self.condition
        # value is empty if condition doesn't match
        # otherwise value is arg
        if not self.is_condition_matches(condition_value):
            value = []
        else:
            value = self.get_args()
        if min_args_len and not len(value) >= min_args_len:
            value = []
        if not value:
            # default value
            new_default = self.value_from_kwargs(default,
                                                 'default_',
                                                 )
            if new_default is not undefined:
                if not isinstance(new_default, list):
                    value = [new_default]
                else:
                    value = new_default
        return value

    def get_args(self):
        """get all arguments
        """
        return list(self.args)


class CalcValuePropertyHelp(CalcValue):
    """special class to display property error
    """
    def get_name(self):
        """get the condition name
        """
        return self.condition['name']

    def get_indexed_name(self, index: int) -> str:
        """get name for a specified index
        """
        condition_index = self.kwargs.get(f'condition_{index}')
        if condition_index is not None and not isinstance(condition_index, dict):
            raise ValueError(_(f'unexpected condition_{index} must have "todict" argument'))
        return condition_index['name']


    def build_property_message(self,
                               name: str,
                               value: Any,
                               ) -> str:
        """prepare message to display error message if needed
        """
        if not self.reverse_condition:
            msg = _('the value of "{0}" is {1}').format(name, value)
        else:
            msg = _('the value of "{0}" is not {1}').format(name, value)
        return msg

    def get_args(self):
        args = super().get_args()
        action = args[0]
        calculated_expected = self.value_from_kwargs(self.expected,
                                                     'expected_',
                                                     to_dict=True)
        if self.condition is not undefined:
            if 'propertyerror' in self.condition:
                msg = self.condition['propertyerror']
            else:
                name = self.get_name()
                if isinstance(calculated_expected, dict):
                    calc_values = calculated_expected.values()
                else:
                    calc_values = [calculated_expected]
                display_value = display_list([str(val) for val in calc_values],
                                             'or',
                                             add_quote=True)
                msg = self.build_property_message(name, display_value)
        else:
            msgs = []
            for key, value in calculated_expected.items():
                name = self.get_indexed_name(key)
                msgs.append(self.build_property_message(name, f'"{value}"'))
            msg = display_list(msgs, self.condition_operator.lower())
        return [(action, f'"{action}" ({msg})')]


calc_value = CalcValue()
calc_value.__name__ = 'calc_value'  # pylint: disable=attribute-defined-outside-init
# function_waiting_for_dict(calc_value)
calc_value_property_help = CalcValuePropertyHelp()
calc_value_property_help.__name__ = 'calc_value_property_help'  # pylint: disable=attribute-defined-outside-init
function_waiting_for_dict(calc_value_property_help)



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""FilenameOption
"""
from ..i18n import _
from .stroption import StrOption


class FilenameOption(StrOption):
    """represents a choice of a file name
    """
    __slots__ = tuple()
    _type = _('file name')

    def validate(self,
                 value: str,
                 ) -> None:
        super().validate(value)
        if not value.startswith('/'):
            raise ValueError(_('must starts with "/"'))



# -*- coding: utf-8 -*-
# Copyright (C) 2014-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""all official option
"""
from .optiondescription import OptionDescription
from .dynoptiondescription import DynOptionDescription
from .syndynoptiondescription import SynDynOptionDescription, SynDynLeadership
from .leadership import Leadership
from .baseoption import submulti
from .symlinkoption import SymLinkOption
from .syndynoption import SynDynOption
from .option import Option
from .choiceoption import ChoiceOption
from .booloption import BoolOption
from .intoption import IntOption
from .floatoption import FloatOption
from .stroption import StrOption, RegexpOption
from .ipoption import IPOption
from .portoption import PortOption
from .networkoption import NetworkOption
from .netmaskoption import NetmaskOption
from .broadcastoption import BroadcastOption
from .domainnameoption import DomainnameOption
from .emailoption import EmailOption
from .urloption import URLOption
from .usernameoption import UsernameOption, GroupnameOption
from .dateoption import DateOption
from .filenameoption import FilenameOption
from .passwordoption import PasswordOption
from .macoption import MACOption
from .permissionsoption import PermissionsOption


__all__ = ('Leadership', 'OptionDescription', 'DynOptionDescription',
           'SynDynOptionDescription', 'SynDynLeadership', 'Option', 'SymLinkOption',
           'SynDynOption', 'ChoiceOption', 'BoolOption', 'DateOption',
           'IntOption', 'FloatOption', 'StrOption',
           'IPOption', 'PortOption', 'NetworkOption', 'NetmaskOption',
           'BroadcastOption', 'DomainnameOption', 'EmailOption', 'URLOption',
           'UsernameOption', 'GroupnameOption', 'FilenameOption', 'PasswordOption', 'submulti',
           'RegexpOption', 'MACOption', 'PermissionsOption')



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""DomainnameOption
"""
import re
from ipaddress import ip_interface
from typing import Any, Optional, List

from ..i18n import _
from ..setting import undefined
from .ipoption import IPOption
from .stroption import StrOption
from .networkoption import NetworkOption
from .option import Calculation


class DomainnameOption(StrOption):
    """represents the choice of a domain name
    netbios: for MS domain
    hostname: to identify the device
    domainname:
    fqdn: with tld, not supported yet
    """
    __slots__ = tuple()
    _type = _('domain name')

    def __init__(self,
                 name: str,
                 doc: str,
                 default: Any=undefined,
                 default_multi: Any=None,
                 multi: bool=False,
                 validators: Optional[List[Calculation]]=None,
                 properties: Optional[List[str]]=None,
                 warnings_only: bool=False,
                 allow_ip: bool=False,
                 allow_cidr_network: bool=False,
                 type: str='domainname',
                 allow_without_dot: bool=False,
                 allow_startswith_dot: bool=False,
                 ) -> None:
        # pylint: disable=too-many-branches,too-many-locals,too-many-arguments
        if type not in ['netbios', 'hostname', 'domainname']:
            raise ValueError(_('unknown type {0} for hostname').format(type))
        extra = {'_dom_type': type}
        if not isinstance(allow_ip, bool):
            raise ValueError(_('allow_ip must be a boolean'))
        if not isinstance(allow_cidr_network, bool):
            raise ValueError(_('allow_cidr_network must be a boolean'))
        if not isinstance(allow_without_dot, bool):
            raise ValueError(_('allow_without_dot must be a boolean'))
        if not isinstance(allow_startswith_dot, bool):
            raise ValueError(_('allow_startswith_dot must be a boolean'))
        extra['_allow_without_dot'] = allow_without_dot
        if type == 'domainname':
            if allow_without_dot:
                min_time = 0
            else:
                min_time = 1
            regexp = r'((?!-)[a-z0-9-]{{{1},{0}}}\.){{{1},}}[a-z0-9-]{{1,{0}}}'.format(self._get_len(type), min_time)
        else:
            regexp = r'((?!-)[a-z0-9-]{{1,{0}}})'.format(self._get_len(type))
        msg = _('must start with lowercase characters followed by lowercase characters, number, "-" and "." characters are allowed')
        msg_warning = _('must start with lowercase characters followed by lowercase characters, number, "-" and "." characters are recommanded')
        if allow_ip:
            msg = _('could be a IP, otherwise {}').format(msg)
            msg_warning = _('could be a IP, otherwise {}').format(msg_warning)
            if not allow_cidr_network:
                regexp = r'(?:{0}|(?:(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){{3}}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))'.format(regexp)
            else:
                regexp = r'(?:{0}|(?:(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){{3}}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)/[0-9][0-9]))'.format(regexp)
        regexp = r'^{0}$'.format(regexp)
        extra['_domain_re'] = re.compile(regexp)
        extra['_domain_re_message'] = msg
        extra['_domain_re_message_warning'] = msg_warning
        extra['_has_upper'] = re.compile('[A-Z]')
        if allow_ip:
            extra['_ip'] = IPOption(name,
                                    doc)
        extra['_allow_ip'] = allow_ip
        if allow_cidr_network:
            extra['_network'] = NetworkOption(name,
                                              doc,
                                              cidr=True)
        extra['_allow_cidr_network'] = allow_cidr_network
        extra['_allow_startswith_dot'] = allow_startswith_dot

        super().__init__(name,
                         doc,
                         default=default,
                         default_multi=default_multi,
                         multi=multi,
                         validators=validators,
                         properties=properties,
                         warnings_only=warnings_only,
                         extra=extra)

    def _get_len(self, type_):
        if type_ == 'netbios':
            return 15
        return 63

    def _validate_domain(self,
                         value: str) -> None:
        def _valid_length(val):
            if len(val) < 1:
                raise ValueError(_("invalid length (min 1)"))
            if len(val) > part_name_length:
                raise ValueError(_("invalid length (max {0})"
                                    "").format(part_name_length))

        part_name_length = self._get_len(self.impl_get_extra('_dom_type'))
        if self.impl_get_extra('_dom_type') == 'domainname':
            if not self.impl_get_extra('_allow_without_dot') and not "." in value:
                raise ValueError(_("must have dot"))
            if len(value) > 255:
                raise ValueError(_("invalid length (max 255)"))
            if self.impl_get_extra('_allow_startswith_dot') and value.startswith('.'):
                val = value[1:]
            else:
                val = value
            if val.endswith('.'):
                nval = val[:-1]
            else:
                nval = val
            for dom in nval.split('.'):
                _valid_length(dom)
        else:
            _valid_length(value)

    def _validate_ip_network(self,
                             value: str) -> None:
        allow_ip = self.impl_get_extra('_allow_ip')
        allow_cidr_network = self.impl_get_extra('_allow_cidr_network')
        if allow_ip is False and allow_cidr_network is False:
            raise ValueError(_('must not be an IP'))
        if allow_ip is True:
            try:
                self.impl_get_extra('_ip').validate(value)
                return
            except ValueError as err:
                if allow_cidr_network is False:
                    raise err
        if allow_cidr_network is True:
            self.impl_get_extra('_network').validate(value)

    def validate(self,
                 value: str) -> None:
        super().validate(value)
        try:
            # check if it's an IP or network
            ip_interface(value)
        except ValueError:
            self._validate_domain(value)
        else:
            self._validate_ip_network(value)

    def _second_level_validation_domain(self,
                                        value: str,
                                        warnings_only: bool) -> None:
        if self.impl_get_extra('_has_upper').search(value):
            raise ValueError(_('some characters are uppercase'))
        if self.impl_get_extra('_allow_startswith_dot') and value.startswith('.'):
            val = value[1:]
        else:
            val = value
        if val.endswith('.'):
            nval = val[:-1]
        else:
            nval = val
        if not self.impl_get_extra('_domain_re').search(nval):
            if warnings_only:
                raise ValueError(self.impl_get_extra('_domain_re_message_warning'))
            raise ValueError(self.impl_get_extra('_domain_re_message'))

    def _second_level_validation_ip_network(self,
                                            value: str,
                                            warnings_only: bool) -> None:
        allow_ip = self.impl_get_extra('_allow_ip')
        allow_cidr_network = self.impl_get_extra('_allow_cidr_network')
        # it's an IP so validate with IPOption
        if allow_ip is True and allow_cidr_network is False:
            try:
                self.impl_get_extra('_ip').second_level_validation(value, warnings_only)
                return
            except ValueError as err:
                raise err
        if allow_cidr_network is True:
            self.impl_get_extra('_network').second_level_validation(value, warnings_only)

    def second_level_validation(self,
                                value: str,
                                warnings_only: bool) -> None:
        try:
            # check if it's an IP or network
            ip_interface(value)
        except ValueError:
            self._second_level_validation_domain(value, warnings_only)
        else:
            self._second_level_validation_ip_network(value, warnings_only)



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""PortOption
"""
import re

from ..i18n import _
from .stroption import StrOption


class PortOption(StrOption):
    """represents the choice of a port
    The port numbers are divided into three ranges:
    the well-known ports,
    the registered ports,
    and the dynamic or private ports.
    You can actived this three range.
    Port number 0 is reserved and can't be used.
    see: http://en.wikipedia.org/wiki/Port_numbers
    """
    __slots__ = tuple()
    port_re = re.compile(r"^[0-9]*$")
    _port = _('port')

    def __init__(self,
                 *args,
                 allow_range: bool=False,
                 allow_zero: bool=False,
                 allow_wellknown: bool=True,
                 allow_registred: bool=True,
                 allow_protocol: bool=False,
                 allow_private: bool=False,
                 **kwargs) -> None:

        extra = {'_allow_range': allow_range,
                 '_allow_protocol': allow_protocol,
                 '_min_value': None,
                 '_max_value': None,
                 }
        ports_min = [0, 1, 1024, 49152]
        ports_max = [0, 1023, 49151, 65535]
        is_finally = False
        for index, allowed in enumerate([allow_zero,
                                         allow_wellknown,
                                         allow_registred,
                                         allow_private]):
            if extra['_min_value'] is None:
                if allowed:
                    extra['_min_value'] = ports_min[index]
            elif not allowed:
                is_finally = True
            elif allowed and is_finally:
                raise ValueError(_('inconsistency in allowed range'))
            if allowed:
                extra['_max_value'] = ports_max[index]

        if extra['_max_value'] is None:
            raise ValueError(_('max value is empty'))

        super().__init__(*args,
                         extra=extra,
                         **kwargs)

    def validate(self,
                 value: str) -> None:
        super().validate(value)
        if self.impl_get_extra('_allow_protocol') and (value.startswith('tcp:') or
                                                       value.startswith('udp:')):
            value = [value[4:]]
        elif self.impl_get_extra('_allow_range') and ":" in str(value):
            value = value.split(':')
            if len(value) != 2:
                raise ValueError(_('range must have two values only'))
            if not value[0] < value[1]:
                raise ValueError(_('first port in range must be'
                                   ' smaller than the second one'))
        else:
            value = [value]

        for val in value:
            if not self.port_re.search(val):
                raise ValueError()

    def second_level_validation(self,
                                value: str,
                                warnings_only: bool) -> None:
        if self.impl_get_extra('_allow_protocol') and (value.startswith('tcp:') or
                                                       value.startswith('udp:')):
            value = [value[4:]]
        elif ':' in value:
            value = value.split(':')
        else:
            value = [value]
        for val in value:
            val = int(val)
            if not self.impl_get_extra('_min_value') <= val <= self.impl_get_extra('_max_value'):
                if warnings_only:
                    msg = 'should be between {0} and {1}'
                else:
                    msg = 'must be between {0} and {1}'
                raise ValueError(_(msg).format(self.impl_get_extra('_min_value'),
                                               self.impl_get_extra('_max_value')))



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""URLOption to check url value
"""
import re
from typing import Any, Optional, List, Dict

from ..setting import undefined
from ..i18n import _
from .option import Calculation
from .stroption import StrOption
from .domainnameoption import DomainnameOption
from .portoption import PortOption


class URLOption(StrOption):
    """URLOption to check url value
    """
    __slots__ = tuple()
    path_re = re.compile(r"^[A-Za-z0-9\-\._~:/\?#\[\]@!%\$&\'\(\)\*\+,;=]+$")
    _type = _('URL')

    def __init__(self,
                 name: str,
                 doc: str,
                 default: Any=undefined,
                 default_multi: Any=None,
                 multi: bool=False,
                 validators: Optional[List[Calculation]]=None,
                 properties: Optional[List[str]]=None,
                 warnings_only: bool=False,
                 extra: Optional[Dict]=None,
                 allow_ip: bool=False,
                 type: str='domainname',
                 allow_without_dot=False,
                 allow_range: bool=False,
                 allow_zero: bool=False,
                 allow_wellknown: bool=True,
                 allow_registred: bool=True,
                 allow_private: bool=False) -> None:
        # pylint: disable=too-many-arguments,too-many-locals,redefined-builtin
        extra = {'_domainname': DomainnameOption(name,
                                                 doc,
                                                 allow_ip=allow_ip,
                                                 type=type,
                                                 allow_without_dot=allow_without_dot),
                 '_port': PortOption(name,
                                     doc,
                                     allow_range=allow_range,
                                     allow_zero=allow_zero,
                                     allow_wellknown=allow_wellknown,
                                     allow_registred=allow_registred,
                                     allow_private=allow_private)}
        super().__init__(name,
                         doc,
                         default=default,
                         default_multi=default_multi,
                         multi=multi,
                         validators=validators,
                         properties=properties,
                         warnings_only=warnings_only,
                         extra=extra)

    def _get_domain_port_files(self,
                               value: str) -> (str, str):
        if value.startswith('http://'):
            type_ = 'http'
            value = value[7:]
        elif value.startswith('https://'):
            type_ = 'https'
            value = value[8:]
        else:
            raise ValueError(_('must start with http:// or '
                                'https://'))
        # get domain/files
        splitted = value.split('/', 1)
        if len(splitted) == 1:
            domain = value
            files = None
        else:
            domain, files = splitted
        # if port in domain
        splitted = domain.split(':', 1)
        if len(splitted) == 1:
            domain = splitted[0]
            port = {'http': '80',
                    'https': '443'}[type_]
        else:
            domain, port = splitted
        return domain, port, files

    def validate(self,
                 value: str) -> None:
        super().validate(value)
        domain, port, files = self._get_domain_port_files(value)
        # validate port
        portoption = self.impl_get_extra('_port')
        portoption.validate(port)
        # validate domainname
        domainnameoption = self.impl_get_extra('_domainname')
        domainnameoption.validate(domain)
        # validate files
        if files is not None and files != '' and not self.path_re.search(files):
            raise ValueError(_('must ends with a valid resource name'))

    def second_level_validation(self, value, warnings_only):
        domain, port, _ = self._get_domain_port_files(value)
        # validate port
        portoption = self.impl_get_extra('_port')
        portoption.second_level_validation(port, warnings_only)
        # validate domainname
        domainnameoption = self.impl_get_extra('_domainname')
        domainnameoption.second_level_validation(domain, warnings_only)



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""ChoiceOption
"""
from typing import Any

from ..setting import undefined, OptionBag
from ..i18n import _
from .option import Option
from ..autolib import Calculation
from ..error import ConfigError, display_list


class ChoiceOption(Option):
    """represents a choice out of several objects.

    The option can also have the value ``None``
    """
    __slots__ = tuple()
    _type = _('choice')

    def __init__(self,
                 name,
                 doc,
                 values,
                 *args,
                 **kwargs):

        """
        :param values: is a list of values the option can possibly take
        """
        if not isinstance(values, (Calculation, tuple)):
            raise TypeError(_('values must be a tuple or a calculation for {0}'
                             ).format(name))
        self._choice_values = values
        super().__init__(name,
                         doc,
                         *args,
                         **kwargs)

    def impl_get_values(self,
                        option_bag: OptionBag,
                        ):
        """get values allowed by option
        """
        if isinstance(self._choice_values, Calculation):
            values = self._choice_values.execute(option_bag)
            if values is not undefined and not isinstance(values, list):
                raise ConfigError(_('the calculated values "{0}" for "{1}" is not a list'
                                    '').format(values, self.impl_getname()))
        else:
            values = self._choice_values
        return values

    def validate(self,
                 value: Any,
                 ) -> None:
        """nothing to valide
        """

    def validate_with_option(self,
                             value: Any,
                             option_bag: OptionBag,
                             loaded: bool,
                             ) -> None:
        if loaded and isinstance(self._choice_values, Calculation):
            return
        values = self.impl_get_values(option_bag)
        self.validate_values(value, values)

    def validate_values(self,
                        value,
                        values,
                        ) -> None:
        """validate values
        """
        if values is not undefined and value not in values:
            if len(values) == 1:
                raise ValueError(_('only "{0}" is allowed'
                                   '').format(values[0]))
            raise ValueError(_('only {0} are allowed'
                               '').format(display_list(values, add_quote=True)))
